"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ComponentParser = void 0;
const change_case_1 = require("change-case");
const path_1 = require("path");
const ELEMENT_PREFIX = 'el';
const CODE_FILE = 'codeFile';
const CSS_FILE = 'cssFile';
const MULTI_TEMPLATE_KEY_ATTRIBUTE = 'key';
const KNOWN_TEMPLATE_SUFFIX = 'Template';
const KNOWN_MULTI_TEMPLATE_SUFFIX = 'Templates';
const KNOWN_PLATFORMS = ['android', 'ios', 'desktop'];
var ElementType;
(function (ElementType) {
    ElementType[ElementType["VIEW"] = 0] = "VIEW";
    ElementType[ElementType["COMMON_PROPERTY"] = 1] = "COMMON_PROPERTY";
    ElementType[ElementType["TEMPLATE"] = 2] = "TEMPLATE";
    ElementType[ElementType["TEMPLATE_ARRAY"] = 3] = "TEMPLATE_ARRAY";
})(ElementType || (ElementType = {}));
// Enums are put in specific order as they are concatenated in the end
var ScopeType;
(function (ScopeType) {
    ScopeType[ScopeType["CORE_IMPORTS"] = 0] = "CORE_IMPORTS";
    ScopeType[ScopeType["CUSTOM_IMPORTS"] = 1] = "CUSTOM_IMPORTS";
    ScopeType[ScopeType["CLASS_START"] = 2] = "CLASS_START";
    ScopeType[ScopeType["SLOT_VIEW_TREE"] = 3] = "SLOT_VIEW_TREE";
    ScopeType[ScopeType["VIEW_TREE"] = 4] = "VIEW_TREE";
    ScopeType[ScopeType["CLASS_END"] = 5] = "CLASS_END";
    // Used for getting enum types count
    ScopeType[ScopeType["SCOPE_COUNT"] = 6] = "SCOPE_COUNT";
})(ScopeType || (ScopeType = {}));
var SpecialTags;
(function (SpecialTags) {
    SpecialTags["SLOT"] = "slot";
    SpecialTags["SLOT_CONTENT"] = "slotContent";
    SpecialTags["TEMPLATE"] = "template";
})(SpecialTags || (SpecialTags = {}));
class ComponentParser {
    constructor(moduleRelativePath, platform) {
        this.openTags = new Array();
        this.resolvedRequests = new Array();
        this.codeScopes = new Array(ScopeType.SCOPE_COUNT);
        this.usedNSTags = new Set();
        // Keep counter for the case of platform tags being inside platform tags
        this.unsupportedPlatformTagCount = 0;
        this.moduleDirPath = '';
        this.moduleRelativePath = '';
        this.treeIndex = -1;
        this.isComponentInitialized = false;
        const { dir, ext, name } = (0, path_1.parse)(moduleRelativePath);
        const componentName = (0, change_case_1.pascalCase)(name);
        this.moduleDirPath = dir;
        this.moduleRelativePath = moduleRelativePath.substring(0, moduleRelativePath.length - ext.length);
        this.platform = platform;
        this.currentViewScope = ScopeType.VIEW_TREE;
        this.codeScopes.fill('');
        this.initialize(componentName);
    }
    handleOpenTag(tagName, attributes) {
        var _a, _b;
        // Component root view and its children have already been parsed
        if (this.isComponentInitialized) {
            throw new Error(`Invalid element ${tagName}. Components can only have a single root view`);
        }
        // Platform tags
        if (KNOWN_PLATFORMS.includes(tagName)) {
            if (tagName.toLowerCase() !== this.platform) {
                this.unsupportedPlatformTagCount++;
            }
            return;
        }
        if (this.unsupportedPlatformTagCount > 0) {
            return;
        }
        const openTagInfo = this.openTags[this.openTags.length - 1];
        const isViewProperty = this.isViewProperty(tagName);
        // Handle view property nesting
        if (isViewProperty && (openTagInfo === null || openTagInfo === void 0 ? void 0 : openTagInfo.type) !== ElementType.VIEW) {
            throw new Error(`Property '${tagName}' can only be nested inside a view tag. Parent tag: ${(_a = openTagInfo === null || openTagInfo === void 0 ? void 0 : openTagInfo.tagName) !== null && _a !== void 0 ? _a : 'None'}`);
        }
        if (openTagInfo != null) {
            this.checkOpenTagNestingConditions(openTagInfo, tagName, attributes);
            openTagInfo.hasOpenChildTag = true;
            openTagInfo.nestedTagCount++;
        }
        let newTagInfo = {
            index: -1,
            tagName,
            attributes,
            type: null,
            nestedTagCount: 0,
            childIndices: [],
            isCustomComponent: false,
            isParentForSlots: false,
            hasOpenChildTag: false
        };
        if (isViewProperty) {
            const [parentTagName, tagPropertyName] = tagName.split('.');
            if (openTagInfo != null && openTagInfo.tagName === parentTagName) {
                newTagInfo.index = openTagInfo.index;
                newTagInfo.tagName = parentTagName;
                newTagInfo.propertyName = tagPropertyName;
                if (tagPropertyName.endsWith(KNOWN_TEMPLATE_SUFFIX)) {
                    newTagInfo.type = ElementType.TEMPLATE;
                    this.codeScopes[this.currentViewScope] += `var ${newTagInfo.propertyName}${openTagInfo.index} = () => {`;
                }
                else if (tagPropertyName.endsWith(KNOWN_MULTI_TEMPLATE_SUFFIX)) {
                    newTagInfo.type = ElementType.TEMPLATE_ARRAY;
                    newTagInfo.templateKeys = [];
                }
                else {
                    newTagInfo.type = ElementType.COMMON_PROPERTY;
                }
            }
            else {
                throw new Error(`Property '${tagName}' is not suitable for parent '${openTagInfo.tagName}'`);
            }
        }
        else if (tagName === SpecialTags.SLOT_CONTENT) {
            if (openTagInfo != null) {
                // Switch scope back to view tree once tag is closed
                this.currentViewScope = ScopeType.SLOT_VIEW_TREE;
                openTagInfo.isParentForSlots = true;
            }
            else {
                throw new Error(`Invalid tag '${tagName}'. Tag has no parent`);
            }
        }
        else if (tagName === SpecialTags.TEMPLATE) {
            if (openTagInfo != null) {
                if (openTagInfo.type === ElementType.TEMPLATE_ARRAY) {
                    newTagInfo.index = openTagInfo.index;
                    newTagInfo.type = ElementType.TEMPLATE;
                    const templateIndex = openTagInfo.childIndices.length;
                    // This is necessary for proper string escape
                    const attrValue = MULTI_TEMPLATE_KEY_ATTRIBUTE in attributes ? attributes[MULTI_TEMPLATE_KEY_ATTRIBUTE] : '';
                    this.codeScopes[this.currentViewScope] += `var ${openTagInfo.propertyName}${openTagInfo.index}_${templateIndex} = () => {`;
                    openTagInfo.templateKeys.push(attrValue);
                    openTagInfo.childIndices.push(templateIndex);
                }
                else {
                    // Ignore tag if it's nested inside a single-template property
                    newTagInfo = null;
                }
            }
            else {
                throw new Error('Failed to parse template. No parent found');
            }
        }
        else {
            const isSlotFallback = (openTagInfo === null || openTagInfo === void 0 ? void 0 : openTagInfo.tagName) === SpecialTags.SLOT;
            if (isSlotFallback) {
                this.codeScopes[this.currentViewScope] += '} else {';
            }
            else {
                this.treeIndex++;
            }
            newTagInfo.index = this.treeIndex;
            newTagInfo.type = ElementType.VIEW;
            if (openTagInfo != null) {
                if (openTagInfo.isParentForSlots) {
                    const slotName = attributes.slot;
                    if (!openTagInfo.slotNames.includes(slotName)) {
                        this.codeScopes[this.currentViewScope] += `slotViews${openTagInfo.index}['${slotName}'] = [];`;
                        openTagInfo.slotNames.push(slotName);
                    }
                }
                else {
                    openTagInfo.childIndices.push(this.treeIndex);
                }
            }
            else {
                // Resolve js and css based on first element
                this.bindCodeAndStyleModules(attributes);
            }
            if (tagName === SpecialTags.SLOT) {
                const name = (_b = attributes.name) !== null && _b !== void 0 ? _b : '';
                this.codeScopes[this.currentViewScope] += `var ${ELEMENT_PREFIX}${this.treeIndex};
        if (this.__slotViews['${name}']) {
          ${ELEMENT_PREFIX}${this.treeIndex} = this.__slotViews['${name}'];`;
            }
            else {
                const [elementName, prefix] = this.getLocalAndPrefixByName(tagName);
                this.buildComponent(elementName, prefix, attributes, isSlotFallback);
                if (prefix != null) {
                    newTagInfo.isCustomComponent = true;
                    newTagInfo.slotNames = [];
                    // A prefixed element has support for nesting views with slot name
                    this.codeScopes[ScopeType.SLOT_VIEW_TREE] += `var slotViews${this.treeIndex} = {};`;
                }
                else {
                    // Store tags that are actually nativescript core components
                    this.usedNSTags.add(tagName);
                }
            }
            if (openTagInfo != null && openTagInfo.isParentForSlots) {
                this.codeScopes[this.currentViewScope] += `slotViews${openTagInfo.index}['${attributes.slot}'].push(${ELEMENT_PREFIX}${this.treeIndex});`;
            }
        }
        newTagInfo != null && this.openTags.push(newTagInfo);
    }
    handleCloseTag(tagName) {
        // Platform tags
        if (KNOWN_PLATFORMS.includes(tagName)) {
            if (tagName.toLowerCase() !== this.platform) {
                this.unsupportedPlatformTagCount--;
            }
            return;
        }
        if (this.unsupportedPlatformTagCount > 0) {
            return;
        }
        let openTagInfo = this.openTags[this.openTags.length - 1];
        if (openTagInfo != null) {
            // Current tag is a closing tag
            if (this.isClosingTag(tagName, openTagInfo)) {
                if (openTagInfo.propertyName != null) {
                    switch (openTagInfo.type) {
                        case ElementType.COMMON_PROPERTY: {
                            if (openTagInfo.childIndices.length) {
                                const viewReferences = openTagInfo.childIndices.map(treeIndex => `${ELEMENT_PREFIX}${treeIndex}`);
                                this.codeScopes[this.currentViewScope] += `global.xmlCompiler.addViewsFromBuilder(${ELEMENT_PREFIX}${openTagInfo.index}, [${viewReferences.join(', ')}], '${openTagInfo.propertyName}');`;
                            }
                            break;
                        }
                        case ElementType.TEMPLATE: {
                            const childIndex = openTagInfo.childIndices[openTagInfo.childIndices.length - 1];
                            this.codeScopes[this.currentViewScope] += `return ${childIndex != null ? ELEMENT_PREFIX + childIndex : 'null'}; };`;
                            this.codeScopes[this.currentViewScope] += `${ELEMENT_PREFIX}${openTagInfo.index}.${openTagInfo.propertyName} = ${openTagInfo.propertyName}${openTagInfo.index};`;
                            break;
                        }
                        case ElementType.TEMPLATE_ARRAY: {
                            const keyedTemplates = openTagInfo.childIndices.map((templateIndex, index) => `{
                ${MULTI_TEMPLATE_KEY_ATTRIBUTE}: '${openTagInfo.templateKeys[index]}',
                createView: ${openTagInfo.propertyName}${openTagInfo.index}_${templateIndex}
              }`);
                            this.codeScopes[this.currentViewScope] += `${ELEMENT_PREFIX}${openTagInfo.index}.${openTagInfo.propertyName} = [${keyedTemplates.join(', ')}];`;
                            break;
                        }
                        default:
                            throw new Error(`Invalid closing property tag '${openTagInfo.tagName}'`);
                    }
                }
                else if (tagName === SpecialTags.SLOT) {
                    this.codeScopes[this.currentViewScope] += '}';
                }
                else if (tagName === SpecialTags.SLOT_CONTENT) {
                    // Get out of slot view tree scope
                    this.currentViewScope = ScopeType.VIEW_TREE;
                }
                else if (tagName === SpecialTags.TEMPLATE) {
                    const childIndex = openTagInfo.childIndices[openTagInfo.childIndices.length - 1];
                    this.codeScopes[this.currentViewScope] += `return ${childIndex != null ? ELEMENT_PREFIX + childIndex : 'null'}; };`;
                }
                else {
                    if (openTagInfo.isParentForSlots && openTagInfo.nestedTagCount) {
                        throw new Error(`Cannot mix common views or properties with slot content inside tag '${tagName}'`);
                    }
                    if (openTagInfo.childIndices.length) {
                        const viewReferences = openTagInfo.childIndices.map(treeIndex => `${ELEMENT_PREFIX}${treeIndex}`);
                        this.codeScopes[this.currentViewScope] += `global.xmlCompiler.addViewsFromBuilder(${ELEMENT_PREFIX}${openTagInfo.index}, [${viewReferences.join(', ')}]);`;
                    }
                }
                // Remove tag from the openTags collection
                this.openTags.pop();
                if (this.openTags.length) {
                    openTagInfo = this.openTags[this.openTags.length - 1];
                }
                else {
                    this.isComponentInitialized = true;
                    openTagInfo = null;
                }
            }
            // Update open child tag flag for current open tag
            if (openTagInfo != null) {
                openTagInfo.hasOpenChildTag = false;
            }
        }
    }
    getResolvedRequests() {
        if (!this.isComponentInitialized) {
            throw new Error('Cannot retrieve required module paths as component parsing has not yet been completed');
        }
        return this.resolvedRequests;
    }
    getResult() {
        if (!this.isComponentInitialized) {
            throw new Error('Cannot retrieve code output as component parsing has not yet been completed');
        }
        let result;
        // Check if component has actual view content
        if (this.treeIndex >= 0) {
            this.appendImportsForUI();
            this.codeScopes[this.currentViewScope] += `resolvedCssModuleName && ${ELEMENT_PREFIX}0.addCssFile(resolvedCssModuleName);
      return ${ELEMENT_PREFIX}0;`;
            result = this.codeScopes.join('');
        }
        else {
            result = '';
        }
        return result;
    }
    initialize(componentName) {
        this.codeScopes[ScopeType.CORE_IMPORTS] += `var { resolveModuleName } = require('@nativescript/core/module-name-resolver');
    var { setPropertyValue } = require('@nativescript/core/ui/builder/component-builder');
    `;
        this.codeScopes[ScopeType.CUSTOM_IMPORTS] += 'var customModules = {};';
        this.codeScopes[ScopeType.CLASS_START] += `export default class ${componentName} {
      constructor(moduleExportsFallback = null) {`;
        this.codeScopes[ScopeType.CLASS_END] += `}
    }
    ${componentName}.isXMLComponent = true;`;
    }
    isClosingTag(closingTagName, openTagInfo) {
        const fullTagName = openTagInfo.propertyName != null ? `${openTagInfo.tagName}.${openTagInfo.propertyName}` : openTagInfo.tagName;
        return fullTagName === closingTagName && !openTagInfo.hasOpenChildTag;
    }
    checkOpenTagNestingConditions(openTagInfo, newTagName, attributes) {
        var _a;
        if (newTagName === SpecialTags.SLOT_CONTENT) {
            if (!openTagInfo.isCustomComponent) {
                throw new Error(`Invalid tag '${newTagName}'. Can only nest slot content inside custom component tags`);
            }
            if (openTagInfo.isParentForSlots) {
                throw new Error(`Invalid tag '${newTagName}'. View already contains a slot content tag`);
            }
            if (openTagInfo.tagName === SpecialTags.SLOT) {
                throw new Error('Cannot nest slot content inside a slot');
            }
        }
        switch (openTagInfo.type) {
            case ElementType.TEMPLATE:
                if (openTagInfo.nestedTagCount) {
                    throw new Error(`Tag '${openTagInfo.tagName}' does not accept more than a single nested element`);
                }
                break;
            case ElementType.TEMPLATE_ARRAY:
                if (newTagName !== SpecialTags.TEMPLATE) {
                    throw new Error(`Property '${openTagInfo.tagName}' must be an array of templates`);
                }
                break;
            default:
                if (openTagInfo.tagName === SpecialTags.SLOT) {
                    if (newTagName === SpecialTags.SLOT) {
                        throw new Error(`Cannot nest slot '${(_a = attributes.slot) !== null && _a !== void 0 ? _a : '<unnamed>'}' inside another slot`);
                    }
                    if (openTagInfo.nestedTagCount) {
                        throw new Error(`Tag '${openTagInfo.tagName}' does not accept more than a single nested element`);
                    }
                }
                if (newTagName === SpecialTags.TEMPLATE) {
                    const fullTagName = openTagInfo.propertyName != null ? `${openTagInfo.tagName}.${openTagInfo.propertyName}` : openTagInfo.tagName;
                    throw new Error(`Template tags can only be nested inside template properties. Parent tag: ${fullTagName}`);
                }
                break;
        }
    }
    appendImportsForUI() {
        if (!this.usedNSTags.size) {
            return;
        }
        const usedTagNames = Array.from(this.usedNSTags).sort();
        this.codeScopes[ScopeType.CORE_IMPORTS] += `var { ${usedTagNames.join(', ')} } = require('@nativescript/core/ui');`;
        this.usedNSTags.clear();
    }
    getPropertyCode(propertyName, propertyValue) {
        let instanceReference = `${ELEMENT_PREFIX}${this.treeIndex}`;
        // This is necessary for proper string escape
        const attrValue = propertyValue.replaceAll('\'', '\\\'');
        if (propertyName.indexOf('.') !== -1) {
            const properties = propertyName.split('.');
            for (let i = 0, length = properties.length - 1; i < length; i++) {
                instanceReference += `?.${properties[i]}`;
            }
            propertyName = properties[properties.length - 1];
        }
        return `${instanceReference} && setPropertyValue(${instanceReference}, null, moduleExports, '${propertyName}', '${attrValue}');`;
    }
    buildComponent(elementName, prefix, attributes, isSlotFallback) {
        let propertyContent = '';
        // Ignore unused attributes
        if ('xmlns' in attributes) {
            delete attributes['xmlns'];
        }
        const entries = Object.entries(attributes);
        for (const [name, value] of entries) {
            if (name === 'slot') {
                continue;
            }
            const [propertyName, prefix] = this.getLocalAndPrefixByName(name);
            if (prefix != null) {
                if (prefix === 'xmlns') {
                    this.registerNamespace(propertyName, value);
                    continue;
                }
                // Platform-based attributes
                if (KNOWN_PLATFORMS.includes(prefix.toLowerCase()) && prefix.toLowerCase() !== this.platform.toLowerCase()) {
                    continue;
                }
            }
            propertyContent += this.getPropertyCode(propertyName, value);
        }
        if (prefix != null) {
            const classRef = `customModules['${prefix}'].${elementName}`;
            this.codeScopes[this.currentViewScope] += `${classRef}.prototype.__slotViews = slotViews${this.treeIndex};`;
            this.codeScopes[this.currentViewScope] += (isSlotFallback ? '' : 'var ') + `${ELEMENT_PREFIX}${this.treeIndex} = ${classRef}.isXMLComponent ? new ${classRef}(moduleExports) : new ${classRef}();`;
            this.codeScopes[this.currentViewScope] += `delete ${classRef}.prototype.__slotViews;`;
        }
        else {
            this.codeScopes[this.currentViewScope] += (isSlotFallback ? '' : 'var ') + `${ELEMENT_PREFIX}${this.treeIndex} = new ${elementName}();`;
        }
        // Apply properties to instance
        this.codeScopes[this.currentViewScope] += propertyContent;
    }
    bindCodeAndStyleModules(attributes) {
        // Script
        if (CODE_FILE in attributes) {
            const attrValue = attributes[CODE_FILE];
            this.resolvedRequests.push(attrValue);
            const resolvedPath = this.getResolvedPath(attrValue);
            this.codeScopes[ScopeType.CLASS_START] += `var resolvedCodeModuleName = resolveModuleName('${resolvedPath}', '');`;
        }
        else {
            this.codeScopes[ScopeType.CLASS_START] += `var resolvedCodeModuleName = resolveModuleName('${this.moduleRelativePath}', '');`;
        }
        // Style
        if (CSS_FILE in attributes) {
            const attrValue = attributes[CSS_FILE];
            this.resolvedRequests.push(attrValue);
            const resolvedPath = this.getResolvedPath(attrValue);
            this.codeScopes[ScopeType.CLASS_START] += `var resolvedCssModuleName = resolveModuleName('${resolvedPath}', 'css');`;
        }
        else {
            this.codeScopes[ScopeType.CLASS_START] += `var resolvedCssModuleName = resolveModuleName('${this.moduleRelativePath}', 'css');`;
        }
        this.codeScopes[ScopeType.CLASS_START] += 'var moduleExports = resolvedCodeModuleName ? global.loadModule(resolvedCodeModuleName, true) : moduleExportsFallback;';
    }
    registerNamespace(propertyName, propertyValue) {
        /**
         * By default, virtual-entry-javascript registers all application js, xml, and css files as modules.
         * Registering namespaces will ensure node modules are also included in module register.
         * However, we have to ensure that the resolved path of files is used as module key so that module-name-resolver works properly.
         */
        this.resolvedRequests.push(propertyValue);
        const resolvedPath = this.getResolvedPath(propertyValue);
        const ext = resolvedPath.endsWith('.xml') ? 'xml' : '';
        // Register module using resolve path as key and overwrite old registration if any
        this.codeScopes[ScopeType.CUSTOM_IMPORTS] += `global.registerModule('${resolvedPath}', () => require('${propertyValue}'));`;
        this.codeScopes[this.currentViewScope] += `customModules['${propertyName}'] = global.xmlCompiler.loadCustomModule('${resolvedPath}', '${ext}');`;
    }
    getLocalAndPrefixByName(name) {
        const splitName = name.split(':');
        let prefix;
        let local;
        if (splitName.length > 1) {
            prefix = splitName[0];
            local = splitName[1];
        }
        else {
            prefix = null;
            local = splitName[0];
        }
        return [
            local,
            prefix
        ];
    }
    getResolvedPath(uri) {
        return uri.startsWith('~/') ? uri.substring(2) : (0, path_1.join)(this.moduleDirPath, uri);
    }
    isViewProperty(tagName) {
        return tagName.indexOf('.') !== -1;
    }
}
exports.ComponentParser = ComponentParser;
//# sourceMappingURL=component-parser.js.map